/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// TODO Fix this warning instead of disabling it
#pragma warning (disable: 4273)


/*
====================================================================================================================================

	NEW FAKE GL WRAPPER - MH 2009

	Uses Direct3D 8 for an (almost) seamless replacement of OpenGL.  Right now it's just restricted to the GL Calls used by
	Quake, but there's no reason why it can't be expanded if need be.

	It is assumed that your OpenGL code is reasonably correct so far as OpenGL is concerned, and therefore a certain layer
	of error checking and validation that *could* be included is actually omitted.  It is recommended that you test any
	OpenGL code you write using native OpenGL calls before sending it through this wrapper in order to ensure this
	correctness.

	Wherever possible the OpenGL specifications at http://www.opengl.org were used for guidance and interpretation of
	OpenGL calls.  The Direct3D 8 SDK help files and several sample aplications and other sources were used for the D3D
	portion of the code.  Every effort has been made to ensure that my interpretation of things is conformant with
	the published specs of each API.  It has been known for published specs to have bugs, and it has also been known
	for my interpretation to be incorrect, however; so let me know if there is anything that needs fixing.

	To use it, add this file and the accompanying "fakegl.h" to your application, and replace your #includes for gl/gl.h and
	gl/glu.h with the same format as is used below.  If you want to also maintain an OpenGL build you can #define the
	REAL_GL_QUAKE macro in your app (or alternatively change this to whatever type of handling you want).

	Compilation should be relatively seamless, although there are some changes you will need to make to your application
	in order to get the best experience from it.  These include the following items (listed with a GLQuake-derived engine
	in mind):

	- call d3dSetMode with the correct params somewhere between VID_SetMode and wglCreateContext (the value of zbpp doesn't
	  matter because we're going to enumerate depth buffer formats properly, so pass in 0.  it's just retained for compatibility
	  with the orginal fakegl wrapper).

	- call FakeSwapBuffers instead of SwapBuffers at the end of each frame.

	- remove any GetProcAddress calls done directly on OpenGL32.dll.

	- #include fakegl.h somewhere where all the c files belonging to the refresh can get at it.

	- remove the #includes for gl/gl.h and gl/glu.h (i suggest the method immediately below)

	Note that I use functions here (such as a replacement for ChangeDisplaySettings) that have Unicode alternatives, but I
	do not provide Unicode support.  If you want to use Unicode you'll really need to do it yourself I'm afraid...

	YOU WILL NEED THE DIRECTX 8 SDK TO COMPILE THIS.

	It's still available on the internet (but not on microsoft.com) so just search on Google.

	It should compile OK on Visual C++ 2008 (including express) but there are a few steps to complete first.

	- Get an unmodified GLQuake compiling and running; there are tutorials on inside3d and/or quakeone for this.

	- Add the DirectX 8 SDK headers and libs to your Visual C++ directories.  I suggest you place them at the very end of the
	  list, as if you add them towards the top some defines or other files may conflict with the version of the platform SDK
	  used for this IDE.

	- create a fake libci.lib static library project (it can just contain one empty function called what you want, compile it,
	  and place the resulting .lib in your source directory so that d3dx8 can link OK.

====================================================================================================================================
*/


#include <windows.h>

// you also need to put this in your glquake.h so that things will link correctly
#ifdef REAL_GL_QUAKE
#include <gl/gl.h>
#include <GL/glu.h>
#else
#include "fakegl.h"
#endif

#include <d3d8.h>
#include <d3dx8.h>

// link to d3d8 libraries this way do that we don't need to modify the project
#pragma comment (lib, "d3d8.lib")
#pragma comment (lib, "d3dx8.lib")

// d3d basic stuff
LPDIRECT3D8 d3d_Object = NULL;
LPDIRECT3DDEVICE8 d3d_Device = NULL;

// externs we need
void Sys_Error (char *error, ...);
void Con_Printf (char *fmt, ...);

// mode definition
HWND d3d_Window;
int d3d_Width;
int d3d_Height;
int d3d_BPP;
BOOL d3d_Windowed;

// the desktop and current display modes
D3DDISPLAYMODE d3d_DesktopMode;
D3DDISPLAYMODE d3d_CurrentMode;

// presentation parameters that the device was created with
D3DPRESENT_PARAMETERS d3d_PresentParams;

// capabilities
// these can exist for both the object and the device
D3DCAPS8 d3d_Caps;

// we're going to need this a lot so store it globally to save me having to redeclare it every time
HRESULT hr = S_OK;

// global state defines
DWORD d3d_ClearColor = 0x00000000;
BOOL d3d_DeviceLost = FALSE;
D3DVIEWPORT8 d3d_Viewport;

// utility
#define BYTE_CLAMP(i) (int) ((((i) > 255) ? 255 : (((i) < 0) ? 0 : (i))))

// go through the vtable rather than use the macros to make this generic
#define SAFE_RELEASE(p) {if (p) (p)->lpVtbl->Release (p); (p) = NULL;}

DWORD GL_ColorToD3D (GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha)
{
	return D3DCOLOR_ARGB
	(
		BYTE_CLAMP (alpha * 255.0f),
		BYTE_CLAMP (red * 255.0f),
		BYTE_CLAMP (green * 255.0f),
		BYTE_CLAMP (blue * 255.0f)
	);
}


/*
===================================================================================================================

			MATRIXES

	Most matrix ops occur in software in D3D; the only part that may (or may not) occur in hardware (depending
	on whether or not we have a hardware vp device) is the final transformation of submitted verts.  Note that
	OpenGL is probably the same.

	This interface just models the most common OpenGL matrix ops with their D3D equivalents.  I don't use the
	IDirect3DXMatrixStack interface because of overhead and baggage associated with it.

	The critical differences are in glGetFloatv and glLoadMatrix; OpenGL uses column major ordering whereas
	D3D uses row major ordering, so we need to translate from one order to the next.

	Also be aware that D3D defines separate world and view matrixes whereas OpenGL concatenates them into a
	single modelview matrix.  To keep things clean and consistent we will always use an identity view matrix
	on a beginscene and will just work off the world matrix.

	Finally, D3D does weird and horrible things in the texture matrix which are better left unsaid for now.

===================================================================================================================
*/

// OpenGL only has a stack depth of 2 for projection, but no reason to not use the full depth
#define MAX_MATRIX_STACK	32

typedef struct d3d_matrix_s
{
	BOOL dirty;
	int stackdepth;
	D3DTRANSFORMSTATETYPE usage;
	D3DMATRIX stack[MAX_MATRIX_STACK];
} d3d_matrix_t;

// init all matrixes to be dirty and at depth 0 (these will be re-inited each frame)
d3d_matrix_t d3d_ModelViewMatrix = {TRUE, 0, D3DTS_WORLD};
d3d_matrix_t d3d_ProjectionMatrix = {TRUE, 0, D3DTS_PROJECTION};

d3d_matrix_t *d3d_CurrentMatrix = &d3d_ModelViewMatrix;


void D3D_InitializeMatrix (d3d_matrix_t *m)
{
	// initializes a matrix to a known state prior to rendering
	m->dirty = TRUE;
	m->stackdepth = 0;
	D3DXMatrixIdentity (&m->stack[0]);
}


void D3D_CheckDirtyMatrix (d3d_matrix_t *m)
{
	if (m->dirty)
	{
		IDirect3DDevice8_SetTransform (d3d_Device, m->usage, &m->stack[m->stackdepth]);
		m->dirty = FALSE;
	}
}


void glMatrixMode (GLenum mode)
{
	switch (mode)
	{
	case GL_MODELVIEW:
		d3d_CurrentMatrix = &d3d_ModelViewMatrix;
		break;

	case GL_PROJECTION:
		d3d_CurrentMatrix = &d3d_ProjectionMatrix;
		break;

	default:
		Sys_Error ("glMatrixMode: unimplimented mode");
		break;
	}
}


void glLoadIdentity (void)
{
	D3DXMatrixIdentity (&d3d_CurrentMatrix->stack[d3d_CurrentMatrix->stackdepth]);
	d3d_CurrentMatrix->dirty = TRUE;
}


void glLoadMatrixf (const GLfloat *m)
{
	memcpy (d3d_CurrentMatrix->stack[d3d_CurrentMatrix->stackdepth].m, m, sizeof (float) * 16);
}


void glFrustum (GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble zNear, GLdouble zFar)
{
	D3DMATRIX tmp;

	// per spec, glFrustum multiplies the current matrix by the specified orthographic projection rather than replacing it
	D3DXMatrixPerspectiveOffCenterRH (&tmp, left, right, bottom, top, zNear, zFar);

	D3DXMatrixMultiply (&d3d_CurrentMatrix->stack[d3d_CurrentMatrix->stackdepth], &tmp, &d3d_CurrentMatrix->stack[d3d_CurrentMatrix->stackdepth]);
	d3d_CurrentMatrix->dirty = TRUE;
}


void glOrtho (GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble zNear, GLdouble zFar)
{
	D3DMATRIX tmp;

	// per spec, glOrtho multiplies the current matrix by the specified orthographic projection rather than replacing it
	D3DXMatrixOrthoOffCenterRH (&tmp, left, right, bottom, top, zNear, zFar);

	D3DXMatrixMultiply (&d3d_CurrentMatrix->stack[d3d_CurrentMatrix->stackdepth], &tmp, &d3d_CurrentMatrix->stack[d3d_CurrentMatrix->stackdepth]);
	d3d_CurrentMatrix->dirty = TRUE;
}


void glPopMatrix (void)
{
	if (!d3d_CurrentMatrix->stackdepth)
	{
		Sys_Error ("glPopMatrix: underflow");
		return;
	}

	// go to a new matrix
	d3d_CurrentMatrix->stackdepth--;

	// flag as dirty
	d3d_CurrentMatrix->dirty = TRUE;
}


void glPushMatrix (void)
{
	if (d3d_CurrentMatrix->stackdepth == (MAX_MATRIX_STACK + 1))
	{
		Sys_Error ("glPushMatrix: overflow");
		return;
	}

	// go to a new matrix
	d3d_CurrentMatrix->stackdepth++;

	// copy up the previous matrix (per spec)
	memcpy
	(
		&d3d_CurrentMatrix->stack[d3d_CurrentMatrix->stackdepth],
		&d3d_CurrentMatrix->stack[d3d_CurrentMatrix->stackdepth - 1],
		sizeof (D3DMATRIX)
	);

	// flag as dirty
	d3d_CurrentMatrix->dirty = TRUE;
}


void glRotatef (GLfloat angle, GLfloat x, GLfloat y, GLfloat z)
{
	// replicates the OpenGL glRotatef with 3 components and angle in degrees
	D3DXVECTOR3 vec;
	D3DMATRIX tmp;

	vec.x = x;
	vec.y = y;
	vec.z = z;

	D3DXMatrixRotationAxis (&tmp, &vec, D3DXToRadian (angle));
	D3DXMatrixMultiply (&d3d_CurrentMatrix->stack[d3d_CurrentMatrix->stackdepth], &tmp, &d3d_CurrentMatrix->stack[d3d_CurrentMatrix->stackdepth]);

	// dirty the matrix
	d3d_CurrentMatrix->dirty = TRUE;
}


void glScalef (GLfloat x, GLfloat y, GLfloat z)
{
	D3DMATRIX tmp;
	D3DXMatrixScaling (&tmp, x, y, z);
	D3DXMatrixMultiply (&d3d_CurrentMatrix->stack[d3d_CurrentMatrix->stackdepth], &tmp, &d3d_CurrentMatrix->stack[d3d_CurrentMatrix->stackdepth]);

	// dirty the matrix
	d3d_CurrentMatrix->dirty = TRUE;
}


void glTranslatef (GLfloat x, GLfloat y, GLfloat z)
{
	D3DMATRIX tmp;
	D3DXMatrixTranslation (&tmp, x, y, z);
	D3DXMatrixMultiply (&d3d_CurrentMatrix->stack[d3d_CurrentMatrix->stackdepth], &tmp, &d3d_CurrentMatrix->stack[d3d_CurrentMatrix->stackdepth]);

	// dirty the matrix
	d3d_CurrentMatrix->dirty = TRUE;
}


/*
===================================================================================================================

			VERTEX SUBMISSION

	Per the spec for glVertex (http://www.opengl.org/sdk/docs/man/xhtml/glVertex.xml) glColor, glNormal and
	glTexCoord just specify a *current* color, normal and texcoord, whereas glVertex actually creates a vertex
	that inherits the current color, normal and texcoord.

	Real D3D code looks *nothing* like this.

===================================================================================================================
*/


int d3d_PrimitiveMode = 0;
int d3d_NumVerts = 0;

D3DCOLOR d3d_CurrentColor = 0xffffffff;
float d3d_CurrentTexCoord[2] = {0, 0};

// this should be a multiple of 12 to support both GL_QUADS and GL_TRIANGLES
// it should also be large enough to hold the biggest tristrip or fan in use in the engine
// individual quads or tris can be submitted in batches
#define D3D_MAX_VERTEXES	600

typedef struct gl_vertex_s
{
	float x, y, z;
	D3DCOLOR c;
	float s, t;
} gl_vertex_t;

gl_vertex_t d3d_Vertexes[D3D_MAX_VERTEXES];

BOOL d3d_SceneBegun = FALSE;

DWORD d3d_FVFCode = D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_TEX1;


void GL_SubmitVertexes (void)
{
	int i;

	// init this to something invalid so that we know it'll never match
	static DWORD d3d_OldFVF = D3DFVF_XYZ | D3DFVF_XYZRHW;

	if (!d3d_Device) return;

	// check for a beginscene
	if (!d3d_SceneBegun)
	{
		// issue a beginscene (geometry needs this
		IDirect3DDevice8_BeginScene (d3d_Device);

		// initial FVF code
		d3d_FVFCode = D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_TEX1;
	}

	if (d3d_FVFCode != d3d_OldFVF || !d3d_SceneBegun)
	{
		// switch FVF (this is also forced after a beginscene
		d3d_OldFVF = d3d_FVFCode;
		IDirect3DDevice8_SetVertexShader (d3d_Device, d3d_FVFCode);
	}

	// we're in a scene now
	d3d_SceneBegun = TRUE;

	// check for dirty states

	// check for dirty matrixes
	D3D_CheckDirtyMatrix (&d3d_ModelViewMatrix);
	D3D_CheckDirtyMatrix (&d3d_ProjectionMatrix);

	// draw the verts - these are the only modes we support for Quake
	switch (d3d_PrimitiveMode)
	{
	case GL_TRIANGLES:
		// D3DPT_TRIANGLELIST models GL_TRIANGLES when used for either a single triangle or multiple triangles
		IDirect3DDevice8_DrawPrimitiveUP (d3d_Device, D3DPT_TRIANGLELIST, d3d_NumVerts / 3, d3d_Vertexes, sizeof (gl_vertex_t));
		break;

	case GL_TRIANGLE_STRIP:
		// regular tristrip
		IDirect3DDevice8_DrawPrimitiveUP (d3d_Device, D3DPT_TRIANGLESTRIP, d3d_NumVerts - 2, d3d_Vertexes, sizeof (gl_vertex_t));
		break;

	case GL_POLYGON:
		// a GL_POLYGON has the same vertex layout and order as a trifan, and can be used interchangably in OpenGL
	case GL_TRIANGLE_FAN:
		// regular trifan
		IDirect3DDevice8_DrawPrimitiveUP (d3d_Device, D3DPT_TRIANGLEFAN, d3d_NumVerts - 2, d3d_Vertexes, sizeof (gl_vertex_t));
		break;

	case GL_QUADS:
		// quads are a special case of trifans where each quad (numverts / 4) represents a trifan with 2 prims in it
		for (i = 0; i < d3d_NumVerts; i += 4)
			IDirect3DDevice8_DrawPrimitiveUP (d3d_Device, D3DPT_TRIANGLEFAN, 2, &d3d_Vertexes[i], sizeof (gl_vertex_t));

		break;

	default:
		// unsupported mode
		break;
	}

	// begin a new primitive
	d3d_NumVerts = 0;
}


void glVertex2f (GLfloat x, GLfloat y)
{
	glVertex3f (x, y, 0);
}


void glVertex3fv (const GLfloat *v)
{
	glVertex3f (v[0], v[1], v[2]);
}


void glVertex3f (GLfloat x, GLfloat y, GLfloat z)
{
	// add a new vertex to the list with the specified xyz and inheriting the current color and texcoords
	// (per spec at http://www.opengl.org/sdk/docs/man/xhtml/glVertex.xml)
	d3d_Vertexes[d3d_NumVerts].x = x;
	d3d_Vertexes[d3d_NumVerts].y = y;
	d3d_Vertexes[d3d_NumVerts].z = z;
	d3d_Vertexes[d3d_NumVerts].c = d3d_CurrentColor;
	d3d_Vertexes[d3d_NumVerts].s = d3d_CurrentTexCoord[0];
	d3d_Vertexes[d3d_NumVerts].t = d3d_CurrentTexCoord[1];

	// go to a new vertex
	d3d_NumVerts++;

	// check for end of vertexes
	if (d3d_NumVerts == D3D_MAX_VERTEXES)
	{
		if (d3d_PrimitiveMode == GL_TRIANGLES || d3d_PrimitiveMode == GL_QUADS)
		{
			// triangles and quads are discrete primitives and as such can begin a new batch
			GL_SubmitVertexes ();
		}
		else
		{
			// other primitives need to extend the vertex storage
		}

		d3d_NumVerts = 0;
	}
}


void glTexCoord2f (GLfloat s, GLfloat t)
{
	d3d_CurrentTexCoord[0] = s;
	d3d_CurrentTexCoord[1] = t;
}


void GL_SetColor (int red, int green, int blue, int alpha)
{
	// overwrite color incase verts set it
	d3d_CurrentColor = D3DCOLOR_ARGB
	(
		BYTE_CLAMP (alpha),
		BYTE_CLAMP (red),
		BYTE_CLAMP (green),
		BYTE_CLAMP (blue)
	);
}


void glColor3f (GLfloat red, GLfloat green, GLfloat blue)
{
	GL_SetColor (red * 255, green * 255, blue * 255, 255);
}


void glColor3ubv (const GLubyte *v)
{
	GL_SetColor (v[0], v[1], v[2], 255);
}


void glColor4f (GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha)
{
	GL_SetColor (red * 255, green * 255, blue * 255, alpha * 255);
}


void glColor4fv (const GLfloat *v)
{
	GL_SetColor (v[0] * 255, v[1] * 255, v[2] * 255, v[3] * 255);
}


void glBegin (GLenum mode)
{
	// just store out the mode, all heavy lifting is done in glEnd
	d3d_PrimitiveMode = mode;

	// begin a new primitive
	d3d_NumVerts = 0;
}


void glEnd (void)
{
	// submit, bitch
	GL_SubmitVertexes ();
}


/*
===================================================================================================================

			TEXTURES

===================================================================================================================
*/

// textures; here we allow space for up to 65536 textures
#define D3D_MAX_TEXTURES	65536

LPDIRECT3DTEXTURE8 d3d_Textures[D3D_MAX_TEXTURES];

// texture binding
int d3d_CurrentTexture = 0;
int d3d_CurrentTMU = 0;

void D3D_ReleaseTextures (void)
{
	int i;

	// explicitly NULL all textures
	for (i = 0; i < D3D_MAX_TEXTURES; i++)
		SAFE_RELEASE (d3d_Textures[i]);
}


// texture creation
// d3d and ogl create textures in very different ways, so here we attempt to meet them inbetween
typedef enum {TCS_NOTSET, TCS_TEXTUREBOUND, TCS_TEXTURETEXIMAGED} texture_create_status_t;

typedef struct texture_create_struct_s
{
	texture_create_status_t createmeatexture;
	BOOL mipmap;
	GLenum target;
	GLint level;
	GLint internalformat;
	GLsizei width;
	GLsizei height;
	GLint border;
	GLenum format;
	GLenum type;
	GLvoid *pixels;
} texture_create_struct_t;

texture_create_struct_t d3d_TextureCreateStruct;


void D3D_CreateTextureFromTextureStruct (void)
{
	D3DLOCKED_RECT lockrect;
	D3DFORMAT texformat = D3DFMT_X8R8G8B8;

	switch (d3d_TextureCreateStruct.internalformat)
	{
	case 1:
	case GL_LUMINANCE:
		texformat = D3DFMT_L8;
		break;

	case 3:
	case GL_RGB:
		texformat = D3DFMT_X8R8G8B8;
		break;

	case 4:
	case GL_RGBA:
		texformat = D3DFMT_A8R8G8B8;
		break;

	default:
		Sys_Error ("invalid texture internal format");
	}

	// create the texture from the data
	hr = IDirect3DDevice8_CreateTexture
	(
		d3d_Device,
		d3d_TextureCreateStruct.width,
		d3d_TextureCreateStruct.height,
		d3d_TextureCreateStruct.mipmap ? 0 : 1,
		0,
		texformat,
		D3DPOOL_MANAGED,
		&d3d_Textures[d3d_CurrentTexture]
	);

	if (FAILED (hr)) Sys_Error ("FAILED (hr) on IDirect3DDevice8_CreateTexture");

	IDirect3DTexture8_LockRect (d3d_Textures[d3d_CurrentTexture], 0, &lockrect, NULL, D3DLOCK_NO_DIRTY_UPDATE);

	if (texformat == D3DFMT_L8)
	{
		int i;
		unsigned char *data = (unsigned char *) lockrect.pBits;

		for (i = 0; i < d3d_TextureCreateStruct.width * d3d_TextureCreateStruct.height; i++)
		{
			data[i] = ((unsigned char *) d3d_TextureCreateStruct.pixels)[i];
		}
	}
	else
	{
		int i;
		unsigned *data = (unsigned *) lockrect.pBits;
		unsigned char *rgba;
		unsigned char tmp;

		for (i = 0; i < d3d_TextureCreateStruct.width * d3d_TextureCreateStruct.height; i++, data++)
		{
			// load and swap to bgra (D3DFMT_A8R8G8B8 is not ARGB, it's BGRA)
			data[0] = ((unsigned *) d3d_TextureCreateStruct.pixels)[i];
			rgba = (unsigned char *) data;
			tmp = rgba[0];
			rgba[0] = rgba[2];
			rgba[2] = tmp;
		}
	}

	IDirect3DTexture8_UnlockRect (d3d_Textures[d3d_CurrentTexture], 0);

	if (d3d_TextureCreateStruct.mipmap)
	{
		// build the mipmap chain
		hr = D3DXFilterTexture ((IDirect3DBaseTexture8 *) d3d_Textures[d3d_CurrentTexture], NULL, 0, D3DX_FILTER_BOX);
		if (FAILED (hr)) Sys_Error ("FAILED (hr) on D3DXFilterTexture");
	}

	// done
	free (d3d_TextureCreateStruct.pixels);
	d3d_TextureCreateStruct.pixels = NULL;

	// flag as dirty
	IDirect3DTexture8_AddDirtyRect (d3d_Textures[d3d_CurrentTexture], NULL);
}


/*
======================
D3D_CheckTextureFormat

Ensures that a given texture format will be available
======================
*/
BOOL D3D_CheckTextureFormat (D3DFORMAT TextureFormat, D3DFORMAT AdapterFormat)
{
	hr = IDirect3D8_CheckDeviceFormat
	(
		d3d_Object,
		D3DADAPTER_DEFAULT,
		D3DDEVTYPE_HAL,
		AdapterFormat,
		0,
		D3DRTYPE_TEXTURE,
		TextureFormat
	);

	return SUCCEEDED (hr);
}


void glTexEnvf (GLenum target, GLenum pname, GLfloat param)
{
	switch ((GLint) param)
	{
	case GL_MODULATE:
		IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_COLOROP, D3DTOP_MODULATE);
		IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_COLORARG1, D3DTA_TEXTURE);
		IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_COLORARG2, D3DTA_DIFFUSE);

		IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_ALPHAOP, D3DTOP_MODULATE);
		IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
		IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE);
		break;

	case GL_REPLACE:
		IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
		IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_COLORARG1, D3DTA_TEXTURE);
		IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_COLORARG2, D3DTA_DIFFUSE);

		IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
		IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
		IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE);
		break;

	default:
		Sys_Error ("glTexEnvf: unimplemented param");
		break;
	}
}


void glTexImage2D (GLenum target, GLint level, GLint internalformat, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const GLvoid *pixels)
{
	// ensure that it's valid to create textures
	if (d3d_TextureCreateStruct.createmeatexture == TCS_NOTSET) return;

	// overwrite an existing texture - just release it so that we can recreate
	SAFE_RELEASE (d3d_Textures[d3d_CurrentTexture]);

	// this is used for both creating new textures and overwriting existing ones
	if (!level)
	{
		int pixelsize = 0;

		// flag that it's been through glTexImage2D and therefore needs to be created
		// this is only valid to be set at level 0
		d3d_TextureCreateStruct.createmeatexture = TCS_TEXTURETEXIMAGED;

		switch (format)
		{
		case 1:
		case GL_LUMINANCE:
			pixelsize = 1;
			break;

		case 4:
		case GL_RGBA:
			pixelsize = 4;
			break;

		default:
			Sys_Error ("glTexImage2D: Unrecognised pixel format");
		}

		switch (type)
		{
		case GL_UNSIGNED_BYTE:
			break;

		default:
			Sys_Error ("glTexImage2D: Unrecognised pixel format");
		}

		// release any pixel memory we might have taken
		if (d3d_TextureCreateStruct.pixels) free (d3d_TextureCreateStruct.pixels);

		// copy the data out to the creation struct
		d3d_TextureCreateStruct.border = border;
		d3d_TextureCreateStruct.format = format;
		d3d_TextureCreateStruct.height = height;
		d3d_TextureCreateStruct.internalformat = internalformat;
		d3d_TextureCreateStruct.level = 0;
		d3d_TextureCreateStruct.pixels = (unsigned char *) malloc (width * height * pixelsize);
		d3d_TextureCreateStruct.target = target;
		d3d_TextureCreateStruct.type = type;
		d3d_TextureCreateStruct.width = width;

		// copy out the data
		memcpy (d3d_TextureCreateStruct.pixels, pixels, width * height * pixelsize);
	}
	else
	{
		// if we call glTexImage2D on a level other than 0 we're creating a mipmap chain
		d3d_TextureCreateStruct.mipmap = TRUE;
	}
}


void glTexParameterf (GLenum target, GLenum pname, GLfloat param)
{
	// fixme - need to check the formal doc on this one!
	IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_ADDRESSU, D3DTADDRESS_WRAP);
	IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_ADDRESSV, D3DTADDRESS_WRAP);
	IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
	IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
	IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_MIPFILTER, D3DTEXF_LINEAR);
}


void glTexSubImage2D (GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels)
{
	int x, y, p;
	D3DLOCKED_RECT lockrect;
	unsigned char *data;

	IDirect3DTexture8_LockRect (d3d_Textures[d3d_CurrentTexture], level, &lockrect, NULL, 0);

	data = (unsigned char *) lockrect.pBits;
	data += yoffset * width + xoffset;

	for (y = yoffset, p = 0; y < (yoffset + height); y++)
	{
		for (x = xoffset; x < (xoffset + width); x++, p++)
		{
			data[p] = ((unsigned char *) pixels)[p];
		}
	}

	IDirect3DTexture8_UnlockRect (d3d_Textures[d3d_CurrentTexture], level);
}


void glBindTexture (GLenum target, GLuint texture)
{
	switch (d3d_TextureCreateStruct.createmeatexture)
	{
	case TCS_TEXTUREBOUND:
		// don't wipe the struct if we're in bound mode
		break;

	case TCS_TEXTURETEXIMAGED:
		// if we're using glBindTexture to create textures we must create the texture now
		D3D_CreateTextureFromTextureStruct ();

		// fall through

	default:
		// we're no longer using glBindTexture to create textures
		d3d_TextureCreateStruct.createmeatexture = TCS_NOTSET;
		d3d_TextureCreateStruct.mipmap = FALSE;

		// free data
		if (d3d_TextureCreateStruct.pixels)
		{
			free (d3d_TextureCreateStruct.pixels);
			d3d_TextureCreateStruct.pixels = NULL;
		}

		break;
	}

	// use no texture
	if (!texture)
	{
		d3d_TextureCreateStruct.createmeatexture = TCS_NOTSET;
		d3d_TextureCreateStruct.mipmap = FALSE;
		IDirect3DDevice8_SetTexture (d3d_Device, d3d_CurrentTMU, NULL);
	}

	// we ignore target
	// a glBindTexture (GL_TEXTURE_2D, 0) is a bind of no texture as textures must start at 1,
	// so we subtract 1 from the texture number to derive the texture to bind
	d3d_CurrentTexture = (texture - 1);

	// OpenGL uses texture binds on textures which have not yet been created so we must check for that
	if (d3d_Textures[d3d_CurrentTexture])
	{
		// here we shove some ugliness in your face; this requires a cast to IDirect3DBaseTexture8 
		IDirect3DDevice8_SetTexture (d3d_Device, d3d_CurrentTMU, (IDirect3DBaseTexture8 *) d3d_Textures[d3d_CurrentTexture]);
	}

	// update the creation struct to reflect that we've bound a texture
	d3d_TextureCreateStruct.createmeatexture = TCS_TEXTUREBOUND;
	d3d_TextureCreateStruct.mipmap = FALSE;
}


/*
===================================================================================================================

			VIDEO SETUP

===================================================================================================================
*/

/*
==================
D3D_GetDepthFormat

Gets a valid depth format for a given adapter format
==================
*/
D3DFORMAT D3D_GetDepthFormat (D3DFORMAT AdapterFormat)
{
	// valid depth formats
	int i;
	D3DFORMAT d3d_DepthFormats[] = {D3DFMT_D32, D3DFMT_D24X8, D3DFMT_D24S8, D3DFMT_D24X4S4, D3DFMT_D16, D3DFMT_UNKNOWN};

	for (i = 0; ; i++)
	{
		if (d3d_DepthFormats[i] == D3DFMT_UNKNOWN)
			break;

		hr = IDirect3D8_CheckDeviceFormat
		(
			d3d_Object,
			D3DADAPTER_DEFAULT,
			D3DDEVTYPE_HAL,
			AdapterFormat,
			D3DUSAGE_DEPTHSTENCIL,
			D3DRTYPE_SURFACE,
			d3d_DepthFormats[i]
		);

		// return the first format that succeeds
		if (SUCCEEDED (hr)) return d3d_DepthFormats[i];
	}

	// didn't get one
	return D3DFMT_UNKNOWN;
}


/*
========================
D3D_GetAdapterModeFormat

returns a usable adapter mode for the given width, height and bpp
========================
*/
D3DFORMAT D3D_GetAdapterModeFormat (int width, int height, int bpp)
{
	int i;

	// fill these in depending on bpp
	D3DFORMAT d3d_Formats[4];

	// these are the orders in which we prefer our formats
	if (bpp == 16)
	{
		d3d_Formats[0] = D3DFMT_R5G6B5;
		d3d_Formats[1] = D3DFMT_X1R5G5B5;
		d3d_Formats[2] = D3DFMT_A1R5G5B5;
		d3d_Formats[3] = D3DFMT_UNKNOWN;
	}
	else
	{
		d3d_Formats[0] = D3DFMT_X8R8G8B8;
		d3d_Formats[1] = D3DFMT_A8R8G8B8;
		d3d_Formats[2] = D3DFMT_UNKNOWN;
	}

	for (i = 0; ; i++)
	{
		UINT modecount;
		UINT m;

		// no more modes
		if (d3d_Formats[i] == D3DFMT_UNKNOWN) break;

		// get and validate the number of modes for this format; we expect that this will succeed first time
		modecount = IDirect3D8_GetAdapterModeCount (d3d_Object, D3DADAPTER_DEFAULT);
		if (!modecount) continue;

		// check each mode in turn to find a match
		for (m = 0; m < modecount; m++)
		{
			// get this mode
			D3DDISPLAYMODE mode;
			hr = IDirect3D8_EnumAdapterModes (d3d_Object, D3DADAPTER_DEFAULT, m, &mode);

			// should never happen
			if (FAILED (hr)) continue;

			// d3d8 doesn't specify a format when enumerating so we need to restrict this to the correct format
			if (mode.Format != d3d_Formats[i]);

			// ensure that we can get a depth buffer
			if (D3D_GetDepthFormat (d3d_Formats[i]) == D3DFMT_UNKNOWN) continue;

			// ensure that the texture formats we want to create exist
			if (!D3D_CheckTextureFormat (D3DFMT_L8, d3d_Formats[i])) continue;
			if (!D3D_CheckTextureFormat (D3DFMT_X8R8G8B8, d3d_Formats[i])) continue;
			if (!D3D_CheckTextureFormat (D3DFMT_A8R8G8B8, d3d_Formats[i])) continue;

			// check it against the requested mode
			if (mode.Width == width && mode.Height == height)
			{
				// copy it out and return the mode we got
				memcpy (&d3d_CurrentMode, &mode, sizeof (D3DDISPLAYMODE));
				return mode.Format;
			}
		}
	}

	// didn't find a format
	return D3DFMT_UNKNOWN;
}


BOOL WINAPI SetPixelFormat (HDC hdc, int format, CONST PIXELFORMATDESCRIPTOR *ppfd)
{
	// just silently pass the PFD through unmodified
	return TRUE;
}


HGLRC WINAPI wglCreateContext (HDC hdc)
{
	// initialize direct3d and get object capabilities
	if (!(d3d_Object = Direct3DCreate8 (D3D_SDK_VERSION)))
		Sys_Error ("Failed to create Direct3D Object");

	if (FAILED (hr = IDirect3D8_GetDeviceCaps (d3d_Object, D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &d3d_Caps)))
		Sys_Error ("failed to get object caps");

	// get the format for the desktop mode
	if (FAILED (hr = IDirect3D8_GetAdapterDisplayMode (d3d_Object, D3DADAPTER_DEFAULT, &d3d_DesktopMode)))
		Sys_Error ("Failed to get desktop mode");

	// clear present params to NULL
	memset (&d3d_PresentParams, 0, sizeof (D3DPRESENT_PARAMETERS));

	if (d3d_Windowed)
	{
		extern int window_width;
		extern int window_height;

		// defaults for windowed mode - also need to store out d3d_Height and d3d_Width
		// (d3d_BPP is only used for fullscreen modes and is retrieved from our CDS override)
		d3d_CurrentMode.Format = d3d_DesktopMode.Format;
		d3d_CurrentMode.Width = d3d_Width = window_width;
		d3d_CurrentMode.Height = d3d_Height = window_height;
		d3d_CurrentMode.RefreshRate = 0;
	}
	else
	{
		// also fills in d3d_CurrentMode
		D3DFORMAT fmt = D3D_GetAdapterModeFormat (d3d_Width, d3d_Height, d3d_BPP);

		// ensure that we got a good format
		if (fmt == D3DFMT_UNKNOWN)
			Sys_Error ("failed to get fullscreen mode");
	}

	// fill in mode-dependent stuff
	d3d_PresentParams.BackBufferFormat = d3d_CurrentMode.Format;
	d3d_PresentParams.FullScreen_RefreshRateInHz = d3d_CurrentMode.RefreshRate;
	d3d_PresentParams.Windowed = d3d_Windowed;

	// request 2 backbuffers
	d3d_PresentParams.BackBufferCount = 2;
	d3d_PresentParams.BackBufferWidth = d3d_Width;
	d3d_PresentParams.BackBufferHeight = d3d_Height;

	d3d_PresentParams.EnableAutoDepthStencil = TRUE;
	d3d_PresentParams.AutoDepthStencilFormat = D3D_GetDepthFormat (d3d_CurrentMode.Format);
	d3d_PresentParams.hDeviceWindow = d3d_Window;
	d3d_PresentParams.MultiSampleType = D3DMULTISAMPLE_NONE;
	d3d_PresentParams.SwapEffect = D3DSWAPEFFECT_DISCARD;

	// here we use D3DCREATE_FPU_PRESERVE to maintain the resolution of Quake's timers (this is a serious problem)
	// and D3DCREATE_DISABLE_DRIVER_MANAGEMENT to protect us from rogue drivers (call it honest paranoia).  first
	// we attempt to create a hardware vp device
	hr = IDirect3D8_CreateDevice
	(
		d3d_Object,
		D3DADAPTER_DEFAULT,
		D3DDEVTYPE_HAL,
		d3d_Window,
		D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_FPU_PRESERVE | D3DCREATE_DISABLE_DRIVER_MANAGEMENT,
		&d3d_PresentParams,
		&d3d_Device
	);

	if (FAILED (hr))
	{
		// it's OK, we may not have hardware vp available, so create a software vp device
		hr = IDirect3D8_CreateDevice
		(
			d3d_Object,
			D3DADAPTER_DEFAULT,
			D3DDEVTYPE_HAL,
			d3d_Window,
			D3DCREATE_SOFTWARE_VERTEXPROCESSING | D3DCREATE_FPU_PRESERVE | D3DCREATE_DISABLE_DRIVER_MANAGEMENT,
			&d3d_PresentParams,
			&d3d_Device
		);

		// oh shit
		if (hr == D3DERR_INVALIDCALL)
			Sys_Error ("IDirect3D8_CreateDevice with D3DERR_INVALIDCALL");
		else if (hr == D3DERR_NOTAVAILABLE)
			Sys_Error ("IDirect3D8_CreateDevice with D3DERR_NOTAVAILABLE");
		else if (hr == D3DERR_OUTOFVIDEOMEMORY)
			Sys_Error ("IDirect3D8_CreateDevice with D3DERR_INVALIDCALL");
		else if (FAILED (hr)) Sys_Error ("failed to create Direct3D device");
	}

	if (d3d_Device == NULL) Sys_Error ("created NULL Direct3D device");

	D3D_ReleaseTextures ();

	// default texture create struct
	d3d_TextureCreateStruct.createmeatexture = TCS_NOTSET;
	d3d_TextureCreateStruct.mipmap = FALSE;
	d3d_TextureCreateStruct.pixels = NULL;

	// disable lighting
	IDirect3DDevice8_SetRenderState (d3d_Device, D3DRS_LIGHTING, FALSE);

	// clear the device on creation
	IDirect3DDevice8_Clear (d3d_Device, 0, NULL, D3DCLEAR_TARGET, 0x00000000, 1, 1);

	return (HGLRC) 1;
}


BOOL WINAPI wglDeleteContext (HGLRC hglrc)
{
	// release all textures
	D3D_ReleaseTextures ();

	// release device and object
	SAFE_RELEASE (d3d_Device);
	SAFE_RELEASE (d3d_Object);

	// success
	return TRUE;
}


HGLRC WINAPI wglGetCurrentContext (VOID)
{
	return (HGLRC) 1;
}


HDC WINAPI wglGetCurrentDC (VOID)
{
	return 0;
}


/*
===================================================================================================================

			COMPARISON TEST AND FUNCTION

	Just DepthFunc and AlphaFunc really

===================================================================================================================
*/

void GL_SetCompFunc (DWORD mode, GLenum func)
{
	switch (func)
	{
	case GL_NEVER:
		IDirect3DDevice8_SetRenderState (d3d_Device, mode, D3DCMP_NEVER);
		break;

	case GL_LESS:
		IDirect3DDevice8_SetRenderState (d3d_Device, mode, D3DCMP_LESS);
		break;

	case GL_LEQUAL:
		IDirect3DDevice8_SetRenderState (d3d_Device, mode, D3DCMP_LESSEQUAL);
		break;

	case GL_EQUAL:
		IDirect3DDevice8_SetRenderState (d3d_Device, mode, D3DCMP_EQUAL);
		break;

	case GL_GREATER:
		IDirect3DDevice8_SetRenderState (d3d_Device, mode, D3DCMP_GREATER);
		break;

	case GL_NOTEQUAL:
		IDirect3DDevice8_SetRenderState (d3d_Device, mode, D3DCMP_NOTEQUAL);
		break;

	case GL_GEQUAL:
		IDirect3DDevice8_SetRenderState (d3d_Device, mode, D3DCMP_GREATEREQUAL);
		break;

	case GL_ALWAYS:
	default:
		IDirect3DDevice8_SetRenderState (d3d_Device, mode, D3DCMP_ALWAYS);
		break;
	}
}


void glDepthFunc (GLenum func)
{
	GL_SetCompFunc (D3DRS_ZFUNC, func);
}


void glAlphaFunc (GLenum func, GLclampf ref)
{
	GL_SetCompFunc (D3DRS_ALPHAFUNC, func);
	IDirect3DDevice8_SetRenderState (d3d_Device, D3DRS_ALPHAREF, BYTE_CLAMP ((int) (ref * 255)));
}


/*
===================================================================================================================

			FRAMEBUFFER

===================================================================================================================
*/

void glDepthRange (GLclampd zNear, GLclampd zFar)
{
	// update the viewport
	d3d_Viewport.MinZ = zNear;
	d3d_Viewport.MaxZ = zFar;

	IDirect3DDevice8_SetViewport (d3d_Device, &d3d_Viewport);
}


void glDepthMask (GLboolean flag)
{
	// if only they were all so easy...
	IDirect3DDevice8_SetRenderState (d3d_Device, D3DRS_ZWRITEENABLE, flag == GL_TRUE ? TRUE : FALSE);
}


void GL_Blend (D3DRENDERSTATETYPE rs, GLenum factor)
{
	D3DBLEND blend = D3DBLEND_ONE;

	switch (factor)
	{
	case GL_ZERO:
		blend = D3DBLEND_ZERO;
		break;

	case GL_ONE:
		blend = D3DBLEND_ONE;
		break;

	case GL_SRC_COLOR:
		blend = D3DBLEND_SRCCOLOR;
		break;

	case GL_ONE_MINUS_SRC_COLOR:
		blend = D3DBLEND_INVSRCCOLOR;
		break;

	case GL_DST_COLOR:
		blend = D3DBLEND_DESTCOLOR;
		break;

	case GL_ONE_MINUS_DST_COLOR:
		blend = D3DBLEND_INVDESTCOLOR;
		break;

	case GL_SRC_ALPHA:
		blend = D3DBLEND_SRCALPHA;
		break;

	case GL_ONE_MINUS_SRC_ALPHA:
		blend = D3DBLEND_INVSRCALPHA;
		break;

	case GL_DST_ALPHA:
		blend = D3DBLEND_DESTALPHA;
		break;

	case GL_ONE_MINUS_DST_ALPHA:
		blend = D3DBLEND_INVDESTALPHA;
		break;

	case GL_SRC_ALPHA_SATURATE:
		blend = D3DBLEND_SRCALPHASAT;
		break;

	default:
		Sys_Error ("glBlendFunc: unknown factor");
	}

	IDirect3DDevice8_SetRenderState (d3d_Device, rs, blend);
}


void glBlendFunc (GLenum sfactor, GLenum dfactor)
{
	GL_Blend (D3DRS_SRCBLEND, sfactor);
	GL_Blend (D3DRS_DESTBLEND, dfactor);
}


void glClear (GLbitfield mask)
{
	DWORD clearflags = 0;

	// no accumulation buffer in d3d
	if (mask & GL_COLOR_BUFFER_BIT) clearflags |= D3DCLEAR_TARGET;
	if (mask & GL_DEPTH_BUFFER_BIT) clearflags |= D3DCLEAR_ZBUFFER;
	if (mask & GL_STENCIL_BUFFER_BIT) clearflags |= D3DCLEAR_STENCIL;

	// glquake hack - always clear depth and colour
	clearflags |= D3DCLEAR_TARGET;
	clearflags |= D3DCLEAR_ZBUFFER;

	// back to normal operation
	IDirect3DDevice8_Clear (d3d_Device, 0, NULL, clearflags, d3d_ClearColor, 1, 1);
}


void glClearColor (GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha)
{
	d3d_ClearColor = GL_ColorToD3D (red, green, blue, alpha);
}


void glDrawBuffer (GLenum mode)
{
	// d3d doesn't like us messing with the front buffer, which is all glquake uses this for, so here
	// we just silently ignore requests to change the draw buffer
}


void glFinish (void)
{
	// we force a Present in our FakeSwapBuffers function so this is unneeded
}


void glReadBuffer (GLenum mode)
{
	// d3d doesn't like us messing with the front buffer, which is all glquake uses this for, so here
	// we just silently ignore requests to change the draw buffer
}


void glReadPixels (GLint x, GLint y, GLsizei width, GLsizei height, GLenum format, GLenum type, GLvoid *pixels)
{
	int row;
	int col;
	unsigned char *srcdata;
	unsigned char *dstdata = (unsigned char *) pixels;
	LPDIRECT3DSURFACE8 bbsurf;
	LPDIRECT3DSURFACE8 locksurf;
	D3DLOCKED_RECT lockrect;
	D3DSURFACE_DESC desc;

	if (format != GL_RGB || type != GL_UNSIGNED_BYTE)
	{
		Con_Printf ("glReadPixels: invalid format or type\n");
		return;
	}

	hr = IDirect3DDevice8_GetBackBuffer (d3d_Device, 0, D3DBACKBUFFER_TYPE_MONO, &bbsurf);

	if (FAILED (hr) || !bbsurf)
	{
		Con_Printf ("glReadPixels: failed to access back buffer\n");
		return;
	}

	// because we don't have a lockable backbuffer we instead copy it off to an image surface
	// this will also handle translation between different backbuffer formats
	hr = IDirect3DSurface8_GetDesc (bbsurf, &desc);
	hr = IDirect3DDevice8_CreateImageSurface (d3d_Device, desc.Width, desc.Height, D3DFMT_R8G8B8, &locksurf);
	hr = D3DXLoadSurfaceFromSurface (locksurf, NULL, NULL, bbsurf, NULL, NULL, D3DX_FILTER_NONE, 0);

	// now we have a surface we can lock
	hr = IDirect3DSurface8_LockRect (locksurf, &lockrect, NULL, D3DLOCK_READONLY);
	srcdata = (unsigned char *) lockrect.pBits;

	// invert the copied image
	for (row = (y + height - 1); row >= y; row--)
	{
		for (col = x; col < (x + width); col++)
		{
			int srcpos = row * width + col;

			// swap to BGR because Quake is going to swap it back
			dstdata[2] = srcdata[srcpos * 3 + 0];
			dstdata[1] = srcdata[srcpos * 3 + 1];
			dstdata[0] = srcdata[srcpos * 3 + 2];

			dstdata += 3;
		}
	}

	hr = IDirect3DSurface8_UnlockRect (locksurf);
	SAFE_RELEASE (locksurf);
	SAFE_RELEASE (bbsurf);
}


void glViewport (GLint x, GLint y, GLsizei width, GLsizei height)
{
	// D3D has a separate view matrix which is concatenated with the world in OpenGL.  We maintain
	// a compatible interface by just setting the D3D view matrix to identity and doing all modelview
	// transforms via the world matrix.
	D3DMATRIX d3d_ViewMatrix;

	// translate from OpenGL bottom-left to D3D top-left
	y = d3d_CurrentMode.Height - (height + y);

	// always take the full depth range
	d3d_Viewport.X = x;
	d3d_Viewport.Y = y;
	d3d_Viewport.Width = width;
	d3d_Viewport.Height = height;
	d3d_Viewport.MinZ = 0;
	d3d_Viewport.MaxZ = 1;

	IDirect3DDevice8_SetViewport (d3d_Device, &d3d_Viewport);

	// set an identity view matrix
	D3DXMatrixIdentity (&d3d_ViewMatrix);
	IDirect3DDevice8_SetTransform (d3d_Device, D3DTS_VIEW, &d3d_ViewMatrix);

	// set projection and world to dirty, beginning of stack and identity
	D3D_InitializeMatrix (&d3d_ModelViewMatrix);
	D3D_InitializeMatrix (&d3d_ProjectionMatrix);
}


void FakeSwapBuffers (void)
{
	// if we lost the device (e.g. on a mode switch, alt-tab, etc) we must try to recover it
	if (d3d_DeviceLost)
	{
		// here we get the current status of the device
		hr = IDirect3DDevice8_TestCooperativeLevel (d3d_Device);

		switch (hr)
		{
		case D3D_OK:
			// device is recovered
			d3d_DeviceLost = FALSE;
			break;

		case D3DERR_DEVICELOST:
			// device is still lost
			break;

		case D3DERR_DEVICENOTRESET:
			// device is ready to be reset
			IDirect3DDevice8_Reset (d3d_Device, &d3d_PresentParams);
			break;
		}

		// don't bother this frame
		return;
	}

	if (d3d_SceneBegun)
	{
		IDirect3DDevice8_EndScene (d3d_Device);
		d3d_SceneBegun = FALSE;
	}

	// present the display
	hr = IDirect3DDevice8_Present (d3d_Device, NULL, NULL, NULL, NULL);

	if (hr == D3DERR_DEVICELOST)
	{
		// flag a lost device
		d3d_DeviceLost = TRUE;
	}
	else if (FAILED (hr))
	{
		// something else bad happened
		Sys_Error ("FAILED (hr) on IDirect3DDevice8_Present");
	}
}


/*
===================================================================================================================

			OTHER

===================================================================================================================
*/

void glCullFace (GLenum mode)
{
	switch (mode)
	{
	case GL_FRONT:
		IDirect3DDevice8_SetRenderState (d3d_Device, D3DRS_CULLMODE, D3DCULL_CW);
		break;

	case GL_BACK:
		IDirect3DDevice8_SetRenderState (d3d_Device, D3DRS_CULLMODE, D3DCULL_CCW);
		break;

	case GL_FRONT_AND_BACK:
	default:
		break;
	}
}


void GL_EnableDisable (GLenum cap, BOOL enabled)
{
	DWORD d3d_RS;

	switch (cap)
	{
	case GL_BLEND:
		d3d_RS = D3DRS_ALPHABLENDENABLE;
		break;

	case GL_ALPHA_TEST:
		d3d_RS = D3DRS_ALPHATESTENABLE;
		break;

	case GL_TEXTURE_2D:
		// this is a texture stage state rather than a render state
		if (enabled)
			IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
		else IDirect3DDevice8_SetTextureStageState (d3d_Device, d3d_CurrentTMU, D3DTSS_COLOROP, D3DTOP_DISABLE);
		return;

	case GL_CULL_FACE:
		d3d_RS = D3DRS_CULLMODE;
		if (enabled) enabled = D3DCULL_CCW; else enabled = D3DCULL_NONE;
		break;

	case GL_FOG:
		return;
		break;

	case GL_DEPTH_TEST:
		d3d_RS = D3DRS_ZENABLE;
		if (enabled) enabled = D3DZB_TRUE; else enabled = D3DZB_FALSE;
		break;

	default:
		return;
	}

	IDirect3DDevice8_SetRenderState (d3d_Device, d3d_RS, enabled);
}


void glDisable (GLenum cap)
{
	GL_EnableDisable (cap, FALSE);
}


void glEnable (GLenum cap)
{
	GL_EnableDisable (cap, TRUE);
}


void glGetFloatv (GLenum pname, GLfloat *params)
{
	switch (pname)
	{
	case GL_MODELVIEW_MATRIX:
		memcpy (params, d3d_ModelViewMatrix.stack[d3d_ModelViewMatrix.stackdepth].m, sizeof (float) * 16);
		break;

	default:
		break;
	}
}


static char *GL_VendorString = "Direct3D8";
static char *GL_RendererString = "Direct3D8";
static char *GL_VersionString = "1.1";
static char *GL_ExtensionString = "Fooled ya!";

const GLubyte *glGetString (GLenum name)
{
	switch (name)
	{
	case GL_VENDOR:
		return GL_VendorString;
	case GL_RENDERER:
		return GL_RendererString;
	case GL_VERSION:
		return GL_VersionString;
	case GL_EXTENSIONS:
	default:
		return GL_ExtensionString;
	}
}


void glHint (GLenum target, GLenum mode)
{
	// ignore driver hints
}


void glPolygonMode (GLenum face, GLenum mode)
{
}


void glShadeModel (GLenum mode)
{
}


void glGetIntegerv (GLenum pname, GLint *params)
{
	// here we only bother getting the values that glquake uses
	switch (pname)
	{
	case GL_MAX_TEXTURE_SIZE:
		// D3D allows both to be different so return the lowest
		params[0] = (d3d_Caps.MaxTextureWidth > d3d_Caps.MaxTextureHeight ? d3d_Caps.MaxTextureHeight : d3d_Caps.MaxTextureWidth);
		break;

	default:
		params[0] = 666;
		return;
	}
}


static void BindTextureExt (GLenum target, GLuint texture)
{
	glBindTexture (target, texture);
}


static void MTexCoord2fSGIS (GLenum target, GLfloat s, GLfloat t)
{
}


static void SelectTextureSGIS (GLenum target)
{
	// do this right...
	d3d_CurrentTMU = 0;
}


// type cast unsafe conversion from
#pragma warning (push)
#pragma warning (disable: 4191)

PROC WINAPI wglGetProcAddress (LPCSTR s)
{
	static LPCSTR kBindTextureEXT = "glBindTextureEXT";
	static LPCSTR kMTexCoord2fSGIS = "glMTexCoord2fSGIS"; // Multitexture
	static LPCSTR kSelectTextureSGIS = "glSelectTextureSGIS";

	if (strncmp (s, kBindTextureEXT, sizeof (kBindTextureEXT) - 1) == 0)
	{
		return (PROC) BindTextureExt;
	}
	else if (strncmp (s, kMTexCoord2fSGIS, sizeof (kMTexCoord2fSGIS) - 1) == 0)
	{
		return (PROC) MTexCoord2fSGIS;
	}
	else if (strncmp (s, kSelectTextureSGIS, sizeof (kSelectTextureSGIS) - 1) == 0)
	{
		return (PROC) SelectTextureSGIS;
	}

	// LocalDebugBreak();
	return 0;
}

#pragma warning (pop)


BOOL WINAPI wglMakeCurrent (HDC hdc, HGLRC hglrc)
{
	return TRUE;
}


void d3dSetMode (int fullscreen, int width, int height, int bpp, int zbpp);
void d3dEvictTextures (void);
void FakeSwapBuffers (void);
void d3dSetGammaRamp (const unsigned char *gammaTable);
void d3dInitSetForce16BitTextures (int force16bitTextures);
void d3dHint_GenerateMipMaps (int value);
float d3dGetD3DDriverVersion ();


void d3dEvictTextures ()
{
}


void d3dSetMode (int fullscreen, int width, int height, int bpp, int zbpp)
{
	extern HWND mainwindow;

	// the value of zpp doesn't matter here; it's just retained for compatibility with the original fakegl wrapper
	// we're going to get a proper z buffer when we create the device.  right now we just store out the vars.
	d3d_Width = width;
	d3d_Height = height;
	d3d_BPP = bpp;
	d3d_Windowed = (fullscreen == 0);
	d3d_Window = mainwindow;
}


void d3dSetGammaRamp (const unsigned char *gammaTable)
{
}


void d3dInitSetForce16BitTextures (int force16bitTextures)
{
}


void d3dHint_GenerateMipMaps (int value)
{
}


float d3dGetD3DDriverVersion (void)
{
	return 8.00f;
}


LONG ChangeDisplaySettings_FakeGL (LPDEVMODE lpDevMode, DWORD dwflags)
{
	// externs to read our properties from
	extern HWND mainwindow;

	// these properties are always set
	d3d_Window = mainwindow;

	if (dwflags & CDS_TEST)
	{
		// if we're testing we want to do it for real
		return ChangeDisplaySettingsA (lpDevMode, dwflags);
	}
	else if (!lpDevMode)
	{
		// this signals a return to the desktop display mode

		// always signal success
		return DISP_CHANGE_SUCCESSFUL;
	}
	else
	{
		// this code path is only used for fullscreen modes
		d3d_Width = lpDevMode->dmPelsWidth;
		d3d_Height = lpDevMode->dmPelsHeight;
		d3d_BPP = lpDevMode->dmBitsPerPel;
		d3d_Windowed = FALSE;

		// always signal success
		return DISP_CHANGE_SUCCESSFUL;
	}
}

